/*
**	File: LSTDIO.C
**
**	QCX Library
**	D.9 --- Input/output
**		D.9.4 Operation on files
**		D.9.5 File access functions
**		D.9.6 Formatted input/output functions 
**		D.9.7 Character input/output functions
**		D.9.8 Direct input/output functions
**		D.9.9 Random-access functions
**		D.9.10 Error-handling functions
**
**	Notes:
**		1. Includes routines which implement common
**		   UNIX(TM) system calls.
**		2. Assumes a CP/M 2.2 target operating system.
**		3. includes the short/long integer only version of
**		   (a) _fmt which supports the printf functions,
**		   (b) _scan which supports scanf/fscanf, and
**		   (c) sscanf.
**		   The floating point versions of all these
**		   routines are in LFLOAT.C
**
**	Copyright (c) 1984, 1985 Quality Computer Systems
**	All rights reserved.
**
**	Published by:	The Code Works
**			Box 62136
**			Santa Barbara, CA 93160
**
**	Modified: 4Apr85 rj
**		 11Sep85 jc
*/

#include "stdio.h"
#include "ctype.h"

/* Library parameters */

#define	TRUE	1
#define	FALSE	0
#define MAXARG	20	/* max no. of command line arguments */
#define MAXLINE 81	/* maximum size of console input line */
#define MAXFLD	128	/* maximum field width for printf formats */
#define MAXSCAN 128	/* maximum size of string in _scanf %[] format */
#define NFILES	10	/* max no. of files that can be open */
#define MAXINT	32767	/* maximum size integer */
#define SECSIZE 128	/* size of CP/M record */

/* Miscellaneous CP/M constants */

#define CPMFCBSIZE 36
#define DBUFF	0x80
#define BUFSIZ	0x200
#define CPMERR	0xFF
#define CPMEOF	0x1A
#define CTLC	0x3	/* ^C */

/* CP/M function numbers */

#define CONIN	1
#define CONOUT	2
#define LSTOUT	5
#define PRINTS	9
#define CONBUF	10
#define CONSTAT 11
#define OPENF	15
#define CLOSEF	16
#define DELETEF 19
#define MAKEF	22
#define SETDMA	26
#define READR	33
#define GTFSIZE 35
#define WRITERZ 40

/* I/O control blocks (iob's) for buffered files
 * struct _iob {
 *	char	_flag;		status flag
 *	char	*_pch;		ptr to curr char in buffer
 *	int	_cnt;		no. of chars remaining in buffer
 *	char	*_buf;		address of buffer
 *	int	_bufsize;	size of buffer
 *	char	_fd;		file descriptor (fd) for low-level I/O
 *	};
 */

/* I/O flag masks */

#define READ	01
#define WRITE	02
#define APPEND	04
#define BINARY	010
#define BUF	020
#define USERBUF 040
#define FEOF	0100
#define FERR	0200

/* File control block for low-level I/O */

struct fcb {
	char	flag;
	unsigned file_size;
	struct cpmfcb {
		char	drive;
		char	filename[8];
		char	filetype[3];
		char	extent;
		int	reserved;
		char	rec_cnt;
		char	dir[16];
		char	curr_rec;
		unsigned rand_rec;
		char	overflow;
		} cpm_fcb;
	} *_fcb[NFILES+3];	/* file control block pointers */

FILE	_iob[NFILES+3] =	/* I/O blocks (iob) for buffered I/O */
		{
		 {READ, NULL, 0, NULL, 0, CONIN},	/* stdin */
		 {WRITE, NULL, 0, NULL, 0, CONOUT},	/* stdout */
		 {WRITE, NULL, 0, NULL, 0, CONOUT}	/* stderr */
		},
	*stdin = &_iob[0],
	*stdout = &_iob[1],
	*stderr = &_iob[2];

int	_ungch = EOF;		/* hold unget character for stdin */

/*>> D.9.4 Operation on files <<*/

/*: remove -- Delete (erase) CP/M file */
remove(filename)
char *filename;
	{
	struct cpmfcb cpmfcb;

	if (makfcb(filename, &cpmfcb) != -1)
		return (bdos1(DELETEF, &cpmfcb) == CPMERR) ? -1: 0;
	return -1;
	}


/*>> D.9.5 File access functions <<*/

/*: fclose -- Close buffered file */
fclose(fp)
register FILE *fp;
	{
	register fd;

	fd = fp->_fd;
	if (fd == CONIN || fd == CONOUT) /* stdin/stdout/stderr */
		return 0;
	if (fd == LSTOUT) {		/* close the printer */
		fp->_flag &= BUF;	/* remember if buffer allocated */
		return 0;
		}
	if (fp->_flag & WRITE		/* if it's an output file */
	   && !(fp->_flag & APPEND)) {	/* not just opened for append */
		if (!(fp->_flag & BINARY)) /* if not binary I/O */
			putc(CPMEOF, fp); /* write a ^Z */
		fflush(fp);		/* write last buffer */
		}
	fp->_flag &= BUF;		/* remember if buffer allocated */
	return close(fd);		/* free file descriptor */
	}

/*: fflush -- Write file buffer */
fflush(fp)
register FILE *fp;
	{
	register nwrite, cnt, fd;
	register char *buf;
	struct fcb *_gfcb();

	fd = fp->_fd;
	if (fd == LSTOUT)		/* to printer? */
		return 0;
	if (!(fp->_flag & WRITE)	/* not open for output? */
	   || _chkbuf(fp) == -1)	/* can't get buffer? */
		fp->_flag |= FERR;
	if (fp->_flag & FERR)
		return EOF;
	if (fp->_flag & APPEND) {	/* position file at CP/M EOF */
		seekr(fd, -1, 2);
		bdos1(SETDMA, buf=fp->_buf);
		bdos1(READR, &((_gfcb(fd))->cpm_fcb));
		bdos1(SETDMA, DBUFF);
		for (cnt = 0; cnt < SECSIZE; ++cnt)
			if (buf[cnt] == CPMEOF)
				break;
		fp->_cnt = cnt; 	/* bytes up to CP/M EOF */
		fp->_pch = buf + cnt;	/* next byte to be written */
		fp->_flag &= ~APPEND;
		return 0;
		}
	if (fp->_cnt == MAXINT) {	/* first time thru */
		fp->_cnt = 0;
		return 0;
		}
	nwrite = ((fp->_cnt + (SECSIZE-1)) / SECSIZE) * SECSIZE;
	fp->_cnt = 0;			/* buffer is empty */
	if (write(fd, fp->_pch=fp->_buf, nwrite) < nwrite) {
		fp->_flag |= FERR;
		return EOF;
		}
	return 0;
	}


/*: fopen -- Open buffered file */
FILE *fopen(filename, mode)
register char *filename;
char *mode;
	{
	register FILE *fp;
	register n, fd, binmode, rwmode;

	for (fp = _iob, n = 0; n < NFILES; ++fp, ++n)
		if (!(fp->_flag & (READ|WRITE|APPEND)))
			break;		/* found empty iob */
	if (n >= NFILES)
		return NULL;		/* no empty iob's */

	binmode = (toupper(*(mode + 1)) == 'B'); /* binary I/O? */
	switch (toupper(*mode)) {
	case 'R':
		rwmode = READ;
		fd = open(filename, 0);
		break;
	case 'W':
		rwmode = WRITE;
		/* check for output to printer (LST:) */
		if (toupper(filename[0]) == 'L' &&
		    toupper(filename[1]) == 'S' &&
		    toupper(filename[2]) == 'T' && filename[3] == ':')
			fd = LSTOUT;
		else
			fd = creat(filename, 0644); /* 0644 = UNIX mode */
		break;
	case 'A':
		rwmode = WRITE;
		if ((fd=open(filename, 1)) == -1)
			fd = creat(filename, 0644);
		else if (binmode)
			seekr(fd, 0, 2); /* start after last rec */
		else
			rwmode = (WRITE|APPEND);
		break;
	default:			/* bad mode */
		return NULL;
		}
	if (fd == -1)
		return NULL;
	if (binmode)
		rwmode |= BINARY;
	fp->_pch = fp->_buf;	/* remember an earlier buffer if any */
	fp->_flag |= rwmode;
	fp->_cnt = (rwmode & WRITE)? MAXINT: 0;
	fp->_fd = fd;
	return fp;
	}

/*: setbuf -- User-provided file buffer */
setbuf(fp, buffer)
FILE *fp;
char *buffer;
	{
	if ((fp->_flag & (BUF|USERBUF)) /* buffer already allocated? */
	    || buffer == NULL)		/* no buffer supplied? */
		return;
	fp->_bufsize = BUFSIZ;
	fp->_pch = fp->_buf = buffer;
	fp->_flag |= USERBUF;
	}


/*>> D.9.6 Formatted input/output functions <<*/

/*: gets -- Get a string from stdin */
char *gets(s)
char *s;
	{
	char *fgets(), *strchr(), *p;

	if (fgets(s, MAXLINE, stdin) == NULL)
		return NULL;
	if (p = strchr(s, '\n')) /* strip newline if present */
		*p = '\0';
	return s;
	}

/*: puts -- Put a string to stdout */
puts(s)
char *s;
	{
	while(*s)
		putchar(*s++);
	putchar('\n');		/* always append newline */
	}

/*: scanf -- Read formatted input from stdin */
scanf(format, args)
char *format, *args;
	{
	return (_scan(stdin, format, &args));
	}

/*: fprintf -- Write formatted output to a file */
fprintf(fp, format, args)
FILE *fp;
char *format, *args;
	{
	int putc();			/* function _fmt will use as output */

	_fmt(putc, fp, format, &args);	/* _fmt does the formatting */
	}

/*: fscanf -- Read formatted input from a file */
fscanf(fp, format, args)
FILE *fp;
char *format, *args;
	{
	return (_scan(fp, format, &args));
	}

/*: _scan -- Internal routine for scanf/fscanf */
_scan(fp, fmt, pargs)
register FILE *fp;
char *fmt;
register char **pargs;
	{
	char ss[MAXSCAN+1], *strchr();
	static char *s;
	static count, c, f, include, present, quit, suppress, width,
		base, d, maxd, longflag, sign;
	static long value;

	count = 0;
	quit = FALSE;
	c = getc(fp);
	while (f = *fmt++) {
		if (isspace(f)) 	/* skip white space in format */
			continue;
		if (f != '%') { 	/* must match next input string */
			while (isspace(c))
				c = getc(fp);
			if (f == c) {
				c = getc(fp); /* matches so get next char */
				continue;
				}
			else
				break;
			}
		if (*fmt == '*') {
			suppress = TRUE;
			++fmt;
			}
		else
			suppress = FALSE;
		width = (isdigit(*fmt))? _atoi(&fmt): -1;
		if (*fmt == 'h') {	/* short int modifier */
			longflag = FALSE;
			++fmt;
			}
		else if (*fmt == 'l') {	/* long integer modifier */
			longflag = TRUE;
			++fmt;
			}
		else
			longflag = FALSE;
		switch (*fmt++) {
		case '%':
			while (isspace(c))
				c = getc(fp);
			if (c == '%') {
				c = getc(fp);
				continue;
				}
			else {
				quit = TRUE;
				break;
				}
		case '[':
			if (*fmt != '^')
				include = TRUE; /* include chars in [] */
			else {
				include = FALSE;
				++fmt;
				}
			for (d = 0; d < MAXSCAN; ++d) {
				f = *fmt++;	/* get char to incl/excl */
				if (f == ']')
					break;
				else if (f == '\0') {
					quit = TRUE;
					break;
					}
				else
					ss[d] = f;
				}
			ss[d] = '\0';
			if (quit)
				break;		/* format didn't end with ] */
			while (isspace(c))
				c = getc(fp);
			if (c == EOF)
				break;
			if (!suppress) {
				++count;
				s = *pargs++;
				}
			while (width--) {
				present = (strchr(ss, c) != NULL);
				if ((include && !present)
				   || (!include && present))
					break;	/* doesn't meet %[s] */
				if (!suppress)
					*s++ = c;
				if ((c = getc(fp)) == EOF)
					break;	/* end of input string */
				}
			if (!suppress)
				*s = '\0';
			break;
		case 'X':
			longflag = TRUE;
		case 'x':
			base = 16;
			maxd = 'F';
			sign = 1;
			goto getnum;
		case 'O':
			longflag = TRUE;
		case 'o':
			base = 8;
			maxd = '7';
			sign = 1;
			goto getnum;
		case 'D':
			longflag = TRUE;
		case 'd':
			base = 10;
			maxd = '9';
			while (isspace(c))
				c = getc(fp);
			if (c == '-') {
				sign = -1;
				c = getc(fp);
				--width;
				}
			else if (c == '+') {
				sign = 1;
				c = getc(fp);
				--width;
				}
			else
				sign = 1;
			goto getnum1;
		getnum:
			while (isspace(c))
				c = getc(fp);
		getnum1:
			d = toupper(c);
			if (isalnum(d) && d <= maxd && width) {
				if (!suppress)
					++count;
				}
			else {
				quit = TRUE;
				break;
				}
			value = 0;
			while (width--) {
				d = toupper(c);
				if (isdigit(d) && d <= maxd)
					d -= '0';
				else if (isalpha(d) && d <= maxd)
					d -= ('A' - 10);
				else
					break;
				/* build negative to handle max negative int */
				value = base * value - d;
				c = getc(fp);
				}
			if (!suppress)
				if (sign > 0)
					value = -value; /* to handle max neg */
				if (longflag)
					*(long *)*pargs++ = value;
				else
					*(int *)*pargs++ = value;
			break;
		case 'C':
		case 'c':
			if (c == EOF)
				break;
			if (suppress) {
				if (width == -1) {
					c = getc(fp); /* get next character */
					continue;
					}
				while (width--) {
					if ((c=getc(fp)) == EOF)
						break;
					}
				break;
				}
			++count;
			if (width == -1) { /* get single char */
				**pargs++ = c;
				c = getc(fp);
				continue;
				}
			else {		/* get string */
				s = *pargs++;
				while (width--) {
					*s++ = c;
					if ((c=getc(fp)) == EOF)
						break;
					}
				*s = '\0';
				}
			break;
		case 'S':
		case 's':
			while (isspace(c))
				c = getc(fp);
			if (c == EOF)
				break;
			if (!suppress) {
				++count;
				s = *pargs++;
				}
			while (width--) {
				if (!suppress)
					*s++ = c;
				c = getc(fp);
				if (isspace(c) || c == EOF)
					break;
				}
			if (!suppress)
				*s = '\0';
			break;
		default:		/* can't figure this out */
			quit = TRUE;
			break;
			}
		if (c == EOF  || quit == TRUE)
			break;
		}
	if (c != EOF)
		ungetc(c, fp);
	return ((count == 0 && c == EOF)? EOF: count);
/* end _scan */	}

/*: printf -- Formatted print routine */
printf(format, args)
char *format, *args;
	{
	int putc();		/* function _fmt will use as output */

	_fmt(putc, stdout, format, &args);
	}

/*: _fmt -- Internal routine for printf/fprintf/sprintf (integer only) */
_fmt(func, dest, format, args)
register (*func)();
register char *dest, **args;
char *format;
{
/*
 * WARNING: the static variables used in this function for compactness
 *	    prevent it from being used in an interrupt environment
 *	    where this routine can be overlaid and then reloaded.
 *	    Change all static declarations to auto or register to
 *	    eliminate this problem.
 */
	static char *sptr;
	register char *wptr;
	register int c;
	static base, pf, ljf, padchr, width, precision;
	int longflag;
	char wbuf[MAXFLD], *itob(), *ltob();

	while (c = *format++) {
	    if (c == '%') {
		wptr = wbuf;
		precision = 6;
		ljf = pf = longflag = FALSE;

		if (*format == '-') {
			++format;
			ljf = TRUE;
			}
		padchr = (*format == '0') ? '0': ' ';
		width = (isdigit(*format)) ? _atoi(&format) : 0;
		if ((c = *format++) == '.') {
			precision = _atoi(&format);
			pf = TRUE;
			c = *format++;
			}
		if (toupper(c) == 'L') {
			longflag = TRUE;
			c = *format++;
			}
		switch(toupper(c)) {

		case 'D':
			base = -10; goto val;
		case 'U':
			base = 10; goto val;
		case 'X':
			base = 16; goto val;
		case 'O':
			base = 8;
		val:
			if (longflag) {
				ltob( *(long *)args, wptr, base);
				args += 2;
				}
			else
				itob( *(int *)args++, wptr, base);
			width -= strlen(wptr);
			goto pad2;
		case 'C':
			*wptr++ = *args++;
			--width;
			goto pad;
		case 'S':
			if (!pf) precision = MAXFLD;
			sptr = *args++;
			while (*sptr && precision) {
				*wptr++ = *sptr++;
				--precision;
				--width;
			}
		pad:	*wptr = '\0';
		pad2:	wptr = wbuf;
			if (!ljf)
				while (--width >= 0)
					(*func)(padchr,dest);
			while (*wptr)
				(*func)(*wptr++,dest);
			if (ljf)
				while (--width >= 0)
					(*func)(padchr,dest);
			break;

		default:
			(*func)(c,dest);
		}
	    }
	    else (*func)(c,dest);
	}
/* end _fmt */ }


/*: ltob -- convert long to string in any base (2-36) */
char *ltob(n, s, base)
long n;
int base;
char *s;
	{
	long u;
	register char *p, *q;
	register negative;
	int c;
	long lundiv();

	if (n < 0 && base < 0) {
		negative = TRUE;
		u = -n;
		}
	else {
		negative = FALSE;
		u = n;
		}
	if (base < 0)	/* signals signed conversion */
		base = -base;
	p = q = s;
	if (u < 0) {
		u = lundiv( u, base, &c);
		if ((*p = c + '0') > '9')
			*p += ('A' - ('9' + 1));
		++p;
		}	/* provides for correct unsigned conversion */
	do {			/* generate digits in reverse order */
		if ((*p = u % base + '0') > '9')
			*p += ('A' - ('9' + 1));
		++p;
		} while ((u /= base) > 0);
	if (negative)
		*p++ = '-';
	*p = '\0';		/* terminate the string */
	while (q < --p) {	/* reverse the digits */
		c = *q;
		*q++ = *p;
		*p = c;
		}
	return s;
/* end ltob */	}
	
/*: lundiv -- divide long int by unsigned int treating long int as unsigned */
long lundiv( n, base, rem)
	long n;
	register unsigned base;
	unsigned *rem;
	{
	register unsigned numerator;
	long result;
	register int count;
	
	numerator = 0;
	result = 0;
	for( count=24; count >= 0; count -= 8) {
		numerator = (numerator<<8) + ((unsigned) ((n>>count) & 255));
		result = (result<<8) + numerator / base;
		numerator %= base;
		}
	*rem = numerator;
	return result;
/* end lundiv */ }

/*: sscanf -- Get formatted input from a string */
sscanf(s, fmt, args)
char *s, *fmt, *args;
	{
	char ss[MAXSCAN+1], *strchr();
	static char *p, **pargs;
	static count, c, f, include, present, quit, suppress, width,
		base, d, maxd, longflag, sign;
	static long value;

	pargs = &args;
	count = 0;
	quit = FALSE;
	c = *s++;
	while (f = *fmt++) {
		if (isspace(f)) 	/* skip white space in format */
			continue;
		if (f != '%') { 	/* must match next input string */
			while (isspace(c))
				c = *s++;
			if (f == c) {
				c = *s++; /* matches so get next char */
				continue;
				}
			else
				break;
			}
		if (*fmt == '*') {
			suppress = TRUE;
			++fmt;
			}
		else
			suppress = FALSE;
		width = (isdigit(*fmt))? _atoi(&fmt): -1;
		if (*fmt == 'h') {	/* short int modifier */
			longflag = FALSE;
			++fmt;
			}
		else if (*fmt == 'l') {	/* long integer modifier */
			longflag = TRUE;
			++fmt;
			}
		else
			longflag = FALSE;
		switch (*fmt++) {
		case '%':
			while (isspace(c))
				c = *s++;
			if (c == '%') {
				c = *s++;
				continue;
				}
			else {
				quit = TRUE;
				break;
				}
		case '[':
			if (*fmt != '^')
				include = TRUE; /* include chars in [] */
			else {
				include = FALSE;
				++fmt;
				}
			for (d = 0; d < MAXSCAN; ++d) {
				f = *fmt++;	/* get char to incl/excl */
				if (f == ']')
					break;
				else if (f == '\0') {
					quit = TRUE;
					break;
					}
				else
					ss[d] = f;
				}
			ss[d] = '\0';
			if (quit)
				break;		/* format didn't end with ] */
			while (isspace(c))
				c = *s++;
			if (c == '\0')
				break;
			if (!suppress) {
				++count;
				s = *pargs++;
				}
			while (width--) {
				present = (strchr(ss, c) != NULL);
				if ((include && !present)
				   || (!include && present))
					break;	/* doesn't meet %[s] */
				if (!suppress)
					*s++ = c;
				if ((c = *s++) == '\0')
					break;	/* end of input string */
				}
			if (!suppress)
				*s = '\0';
			break;
		case 'X':
			longflag = TRUE;
		case 'x':
			base = 16;
			maxd = 'F';
			sign = 1;
			goto getnum;
		case 'O':
			longflag = TRUE;
		case 'o':
			base = 8;
			maxd = '7';
			sign = 1;
			goto getnum;
		case 'D':
			longflag = TRUE;
		case 'd':
			base = 10;
			maxd = '9';
			while (isspace(c))
				c = *s++;
			if (c == '-') {
				sign = -1;
				c = *s++;
				--width;
				}
			else if (c == '+') {
				sign = 1;
				c = *s++;
				--width;
				}
			else
				sign = 1;
			goto getnum1;
		getnum:
			while (isspace(c))
				c = *s++;
		getnum1:
			d = toupper(c);
			if (isalnum(d) && d <= maxd && width) {
				if (!suppress)
					++count;
				}
			else {
				quit = TRUE;
				break;
				}
			value = 0;
			while (width--) {
				d = toupper(c);
				if (isdigit(d) && d <= maxd)
					d -= '0';
				else if (isalpha(d) && d <= maxd)
					d -= ('A' - 10);
				else
					break;
				/* build negative to handle max negative int */
				value = base * value - d;
				c = *s++;
				}
			if (!suppress)
				if (sign > 0)
					value = -value; /* to handle max neg */
				if (longflag)
					*(long *)*pargs++ = value;
				else
					*(int *)*pargs++ = value;
			break;
		case 'C':
		case 'c':
			if (c == '\0')
				break;
			if (suppress) {
				if (width == -1) {
					c = *s++; /* get next character */
					continue;
					}
				while (width--) {
					if ((c=*s++) == '\0')
						break;
					}
				break;
				}
			++count;
			if (width == -1) { /* get single char */
				**pargs++ = c;
				c = *s++;
				continue;
				}
			else {		/* get string */
				p = *pargs++;
				while (width--) {
					*p++ = c;
					if ((c=*s++) == '\0')
						break;
					}
				*p = '\0';
				}
			break;
		case 'S':
		case 's':
			while (isspace(c))
				c = *s++;
			if (c == '\0')
				break;
			if (!suppress) {
				++count;
				p = *pargs++;
				}
			while (width--) {
				if (!suppress)
					*p++ = c;
				c = *s++;
				if (isspace(c) || c == '\0')
					break;
				}
			if (!suppress)
				*p = '\0';
			break;
		default:		/* can't figure this out */
			quit = TRUE;
			break;
			}
		if (c == '\0'  || quit == TRUE)
			break;
		}
	return ((count == 0 && c == '\0')? EOF: count);
/* end sscanf */	}

/*: sprintf -- format a string */
sprintf(s, format, args)
char *s, *format, *args;
	{
	int _cat();			/* function _fmt will use as output */

	*s = '\0';			/* initialize the string */
	_fmt(_cat, &s, format, &args);
	}

/*: _cat -- add character to end of string (internal routine for sprintf) */
_cat(c, s)
char **s;
	{
	*(*s)++ = c;		/* add c at end and advance pointer*/
	*(*s) = '\0';		/* new end of string */
	}

/*: getchar -- Get character from standard input file */
getchar()
	{
	return getc(stdin);
	}

/*: putchar -- Put character to standard output file */
putchar(c)
	{
	return putc(c, stdout);
	}


/*>> D.9.7 Character input/output functions <<*/

/*: fgetc -- Get character from buffered file (true function) */
fgetc(fp)
FILE *fp;
	{
	return getc(fp);
	}

/*: fgets -- Get string from buffered file */
char *fgets(s, maxsize, fp)
char *s;
FILE *fp;
	{
	register int c, size;
	register char *ps, *q;

	ps = s;
	if (fp->_fd == CONIN) { 	/* use CP/M func #10 for console */
		*ps = maxsize;		/* set buffer size for CP/M */
		bdos1(CONBUF, ps);	/* get the line from console */
		bdos1(CONOUT, '\n');	/* echo LF after CR */
		q = ps + 2;		/* beginning of line */
		size = ps[1];		/* find out how long line is */
		while (size--)		/* move line to start of buffer */
			*ps++ = *q++;
		*ps = '\n';		/* append newline */
		*(ps + 1) = '\0';	/* terminate line */
		return (*s == CPMEOF) ? NULL: s;
		}
	while (--maxsize > 0 && (c = getc(fp)) != EOF)
		if ((*ps++ = c) == '\n')
			break;
	*ps = '\0';
	return ((c == EOF && ps == s) ? NULL : s);
	}

/*: fputc -- Put character to buffered file (true function) */
fputc(c, fp)
FILE *fp;
	{
	return putc(c, fp);
	}

/*: fputs -- Put string to buffered file */
fputs(s, fp)
char *s;
FILE *fp;
	{
	register int c;

	while (c = *s++)
		putc(c, fp);
	}

/*: fread -- Read data items from a file */
fread(p, size, nitems, fp)
char *p;
FILE *fp;
	{
	register i, cnt;

	for (cnt = 0; cnt < nitems; ++cnt)
		for (i = 0; i < size; ++i)
			if ((*p++ = getc(fp)) == EOF)
				return (i == 0)? cnt: 0;
	return cnt;
	}

/*: fwrite -- Write data items to a file */
fwrite(p, size, nitems, fp)
char *p;
FILE *fp;
	{
	register i, cnt;

	for (cnt = 0; cnt < nitems; ++cnt) {
		for (i = 0; i < size; ++i) {
			putc(*p++, fp);
			if (ferror(fp))
				return 0;
			}
		}
	return cnt;
	}

/*: getc -- Get character from buffered file */
getc(fp)
register FILE *fp;
	{
	register c;

	if (fp->_fd == CONIN) { 	/* input from console? */
		if ((c = _ungch) != EOF) /* unget char ready? */
			_ungch = EOF;	/* use it and clear */
		else
			if ((c = bdos1(CONIN)) == '\r') /* chg CR */
				bdos1(CONOUT, c = '\n'); /* to newline */
		return (c == CPMEOF) ? EOF: c;
		}
	if (fp->_cnt > 0) {
		c = (*fp->_pch++ & 0xFF); /* defeat sign extension */
		--fp->_cnt;
		}
	else
		c =  _fill(fp);
	if (fp->_flag & BINARY) 	/* no tampering if binary I/O */
		return c;
	if (c == '\r')			/* throw away CR */
		c = getc(fp);
	if (c == CPMEOF) {
		fp->_cnt = 0;
		fp->_flag |= FEOF;
		return EOF;
		}
	return ((c == EOF)? EOF: c & 0x7F);
	}

/*: _fill -- Internal routine to load the buffer for getc */
_fill(fp)
register FILE *fp;	 /* pointer to iob for file */
	{
	register cnt;

	if (fp->_flag & FEOF)
		return EOF;
	cnt = (_chkbuf(fp) == -1) ? -1:
		read(fp->_fd, fp->_buf, fp->_bufsize);
	if (cnt <= 0) { 		/* did we read anything? */
		fp->_flag |= ((cnt==0)? FEOF: FERR);
		return EOF;
		}
	fp->_cnt = cnt - 1;		/* # chars remaining */
	fp->_pch = fp->_buf + 1;	/* point to 2nd char in buffer */
	return (*fp->_buf & 0xFF);	/* return 1st char in buffer */
	}

/*: putc -- Put character to buffered file */
putc(c, fp)
register c;
register FILE *fp;
	{
	if (fp->_fd == CONOUT) {	/* to console? */
		if (c == '\n')
			bdos1(CONOUT, '\r');
		bdos1(CONOUT, c);
		return c;
		}
	if (c == '\n' && !(fp->_flag & BINARY)) /* normal I/O only */
		if (putc('\r', fp) == EOF)
			return EOF;
	if (fp->_fd == LSTOUT) {	/* to printer? */
		bdos1(LSTOUT, c);
		return c;
		}
	if (fp->_cnt >= fp->_bufsize) { /* buffer full? */
		if (fflush(fp) == EOF)
			return EOF;
		}
	++fp->_cnt;
	return (*fp->_pch++ = c);
	}


/*: ungetc -- Push character back onto buffered file */
ungetc(c, fp)
register FILE *fp;
	{
	if (fp->_fd == CONIN) { 	/* stdin */
		if (_ungch != EOF)
			return EOF;	/* only one pushback */
		else
			return (_ungch = c);
		}
	if ((fp->_flag & READ)		/* must be input */
	   && fp->_pch > fp->_buf) {	/* with room to unget */
		++fp->_cnt;
		return (*--fp->_pch = c);
		}
	else
		return EOF;
	}

/*: feof -- Check for end-of-file */
feof(fp)
FILE *fp;
	{
	return (fp->_flag & FEOF);
	}

/*: ferror -- Check file error status */
ferror(fp)
FILE *fp;
	{
	return (fp->_flag & FERR);
	}

/*: clearerr -- Clear file error status */
clearerr(fp)
FILE *fp;
	{
	struct fcb *fcb, *_gfcb();

	fp->_flag &= ~FERR;
	if ((fcb = _gfcb(fp->_fd)) != NULL) /* get low level control block */
		fcb->flag &= ~FERR;	/* clear low level error flag */
	}

/*: open -- Open CP/M file (system call) */
open(filename, rwmode)
char *filename;
register rwmode;
	{
	register struct fcb *fcb;
	register struct cpmfcb *cpmfcb;
	struct fcb *_gfcb();
	register int fd;

	if (rwmode < 0 || rwmode > 2)
		return -1;
	fd = _gfd();			/* get a file descriptor */
	if ((fcb=_gfcb(fd)) == NULL)	/* get file control block for fd */
		return -1;
	cpmfcb = &fcb->cpm_fcb;
	if (makfcb(filename, cpmfcb) == -1
	   || bdos1(OPENF, cpmfcb) == CPMERR)
		return -1;
	fcb->flag = rwmode + 1; 	/* flag: R=01, W=02, R/W=03 */
	bdos1(GTFSIZE, cpmfcb); 	/* get file size */
	fcb->file_size = cpmfcb->rand_rec;
	cpmfcb->rand_rec = 0;
	return fd;
	}

/*: creat -- Create CP/M file (system call) */
creat(filename, pmode) /* p[rotection] mode - ignored by CP/M */
char *filename;
	{
	register struct fcb *fcb;
	register struct cpmfcb *cpmfcb;
	struct fcb *_gfcb();
	register int fd;

	fd = _gfd();			/* get a file descriptor */
	if ((fcb=_gfcb(fd)) == NULL)	/* get file control blk for fd */
		return -1;
	cpmfcb = &fcb->cpm_fcb;
	if (makfcb(filename, cpmfcb) == -1)
		return -1;		/* bad filename */
	bdos1(DELETEF, cpmfcb); 	/* delete any existing file */
	if (bdos1(MAKEF, cpmfcb) == CPMERR)
		return -1;		/* CP/M can't create filename */
	fcb->flag = WRITE;
	fcb->file_size = 0;
	return fd;
	}

/*: read -- Read CP/M logical records (system call) */
read(fd, buf, n)
register n;
char *buf;
	{
	register int i, nread;
	register struct fcb *fcb;
	register struct cpmfcb *cpmfcb;
	struct fcb *_gfcb();

	if ((fcb = _gfcb(fd)) == NULL	/* fd OK? */
	   || (!(fcb->flag & READ))	/* open for input? */
	   || (n % SECSIZE))		/* multiple of CP/M sector? */
		return -1;
	if (fcb->flag & (FEOF|FERR))
		return 0;		/* nothing read - at EOF */
	n /= SECSIZE;			/* convert to CP/M sectors */
	cpmfcb = &fcb->cpm_fcb;
	for (nread = 0; nread < n; ++nread) {
		bdos1(SETDMA, buf);	/* set DMA to our buffer */
		switch (bdos1(READR, cpmfcb)) {
		case 1: 		/* unwritten data */
		case 4: 		/* chk for EOF */
			if (cpmfcb->rand_rec >= fcb->file_size)
				fcb->flag |= FEOF;
			else for (i=0; i<SECSIZE; ++i)
				*buf++ = 0;
			break;
		default:		/* misc errors */
			fcb->flag |= FERR;
			break;
		case 0:
			buf += SECSIZE; /* next sector in our buffer */
			break;
			}
		if (fcb->flag & (FEOF|FERR))
			break;
		++cpmfcb->rand_rec;
		}
	bdos1(SETDMA, DBUFF);		/* reset CP/M default buffer */
	return (nread * SECSIZE);	/* convert to bytes */
	}

/*: write -- Write CP/M logical records (system call) */
write(fd, buf, n)
register n;
char *buf;
	{
	register int nwritten;
	register struct fcb *fcb;
	register struct cpmfcb *cpmfcb;
	struct fcb *_gfcb();

	if ((fcb = _gfcb(fd)) == NULL	/* fd OK? */
	   || (!(fcb->flag & WRITE))	/* open for output? */
	   || n % SECSIZE		/* multiple of CP/M sector? */
	   || fcb->flag & FERR) 	/* previous error? */
		return -1;
	n /= SECSIZE;			/* convert to CP/M sectors */
	cpmfcb = &fcb->cpm_fcb;
	for (nwritten = 0; nwritten < n; ++nwritten) {
		bdos1(SETDMA, buf);	/* set DMA to our buffer */
		if (bdos1(WRITERZ, cpmfcb)) { /* check for error */
			fcb->flag |= FERR;
			break;
			}
		++cpmfcb->rand_rec;
		buf += SECSIZE; 	/* next sector in our buffer */
		}
	if (cpmfcb->rand_rec > fcb->file_size)
		fcb->file_size = cpmfcb->rand_rec;
	bdos1(SETDMA, DBUFF);		/* reset CP/M default buffer */
	return (nwritten * SECSIZE);	/* convert to bytes */
	}

/*: close -- Close CP/M file (system call) */
close(fd)
	{
	register struct fcb *fcb;
	struct fcb *_gfcb();
	register fcbflag;

	if ((fcb = _gfcb(fd)) == NULL)
		return -1;
	fcbflag = fcb->flag;		/* save fcb flag */
	fcb->flag = 0;			/* mark fcb closed */
	if (fcbflag & FERR)		/* has error occurred? */
		return -1;
	if (fcbflag & WRITE)		/* was file open for write? */
		if (bdos1(CLOSEF, &fcb->cpm_fcb) == CPMERR)
			return -1;	/* CP/M can't close */
	return 0;
	}

/*: _gfcb -- Internal routine: get file control block for file descriptor fd */
struct fcb *_gfcb(fd)
register fd;
	{
	char *sbrk();
	register struct fcb *fcb;

	fd -= 6;	/* skip stdin, stdout, stderr, CP/M LST: */
	if (fd >= 0 && fd < NFILES) {
		if ((fcb = _fcb[fd]) == NULL) { /* need space? */
			fcb = (struct fcb *) sbrk(sizeof(struct fcb));
			if ((int)fcb == -1)
				return NULL;
			else {
				fcb->flag = 0;
				_fcb[fd] = fcb;
				}
			}
		return fcb;
		}
	return NULL;
	}

/*: _gfd -- Internal routine to get next available file descriptor */
_gfd()
	{
	register fd;
	struct fcb *_gfcb(), *fcb;

	fd = 6; 	/* skip stdin, stdout, stderr, CP/M LST: */
	while ((fcb = _gfcb(fd)) != NULL) {
		if (fcb->flag == 0)
			return fd;	/* fcb not in use */
		++fd;
		}
	return -1;			/* all fcb's in use */
	}

/*: _chkbuf -- internal routine to check for buffer allocated */
_chkbuf(fp)
register FILE *fp;
	{
	char *sbrk();
	if (!(fp->_flag & (BUF|USERBUF))) { /* buffer allocated? */
		fp->_bufsize = BUFSIZ;
		if ((int)(fp->_pch=fp->_buf=sbrk(BUFSIZ)) == -1) {
			fp->_flag |= FERR;
			return -1;
			}
		else
			fp->_flag |= BUF;
		}
	return 0;
	}

/*: fileno -- Get file descriptor number */
fileno(fp)
FILE *fp;
	{
	return (fp->_fd);
	}

/* end LSTDIO.C */
